#pragma once

#include <pqxx/pqxx>

#include "hack/utils/func_query.hpp"
#include "hack/utils/singleton.hpp"
#include "hack/logger/logger.hpp"
#include "hack/exception/exception.hpp"

#include "trs/utils/var.hpp"
#include "trs/utils/using.hpp"

namespace trs
{
  class database : public hack::utils::singleton<database>
  {
    friend hack::utils::singleton<database>;

    public:
      ~database() = default;

    private:
      database() = default;
      std::mutex m_mutex;

      std::string DB_APP_CONNECTION;
      std::string DB_LOG_CONNECTION;

    public:
      void set_log_connection(std::string con)
      {
        DB_LOG_CONNECTION = con;
      }

      void set_app_connection(std::string con)
      {
        DB_APP_CONNECTION = con;
      }

      template<typename... Args>
      JSON execute(const std::string func_name, const Args... args)
      {
        std::string query;
        JSON result;

        try
        {
          query = hack::utils::make_query(func_name, args...);
          result = execute(func_name, query);
        }
        catch(const hack::exception& ex)
        {
          throw ex;
        }
        catch (const std::exception& e)
        {
          hack::exception ex;
          ex.description("database dont create query from args");
          ex.system_error(e);
          ex.params("query", query);
          ex.variadic_params(args...);
          throw ex; 
        }

        return result;
      }

    private:
      JSON execute(const std::string& func_name, const std::string& query)
      {
        if (DB_APP_CONNECTION.empty() || DB_LOG_CONNECTION.empty())
        {
          hack::error()("DB_LOG_CONNECTION or DB_LOG_CONNECTION is empty");
          hack::log("")("DB_APP_CONNECTION = ", DB_APP_CONNECTION);
          hack::log("")("DB_LOG_CONNECTION = ", DB_LOG_CONNECTION);
          std::terminate();
        }

        JSON result;
        pqxx::connection* app_con;
        pqxx::connection* log_con;

        try
        {
          app_con = new pqxx::connection(DB_APP_CONNECTION);
          log_con = new pqxx::connection(DB_LOG_CONNECTION);
        }
        catch (const std::exception& e)
        {
          hack::error()("Dont create connection");
          hack::error()(e.what());
          std::terminate();
        }
        catch(...)
        {
          hack::error()("Dont create connection and no get error in description");
          hack::error()("Good luck my friend! ...");
          std::terminate();
        }

        pqxx::result r;
        pqxx::nontransaction work { func_name == var::FUNC_LOGGER ? *log_con : *app_con };

        try
        {
          std::lock_guard<std::mutex> lock(m_mutex);
          r = work.exec(query);
          std::string r_str;
          for (auto row : r) r_str = row[func_name].c_str();
          work.commit();
          result = JSON::parse(r_str);
        }
        catch (const std::exception& e)
        {
          if (func_name != var::FUNC_LOGGER)
          {
            hack::exception ex;
            ex.description("No exec query to database");
            ex.message(var::EXECUTE_ERROR);
            ex.system_error(e);
            ex.params("query", query);
            ex.params("result", result);
            throw ex;
          }
          else
          {
            hack::error()("WARNING!!! ERROR LOG TO DATABASE");
            hack::error()("query", query);
            hack::error()(e.what());
            std::terminate();
          }
        }
        catch (...)
        {
          if (func_name != var::FUNC_LOGGER) 
          {
            hack::exception ex;
            ex.description(var::ALIEN_SYSTEM_ERROR);
            ex.message(var::EXECUTE_ERROR);
            ex.params("query", query);
            ex.params("result", result);
            throw ex;
          }
          else
          {
            hack::error()("WARNING!!! ERROR LOG TO DATABASE");
            hack::error()("query", query);
            std::terminate();
          }
        }

        delete app_con;
        delete log_con;

        return result;
      }
  }; 
}
