#pragma once

#include <string>

#include "httplib.h"
#include "hack/logger/logger.hpp"
#include "hack/exception/exception.hpp"

#include "trs/utils/define.hpp"
#include "trs/utils/helpers.hpp"

#include "trs/utils/var.hpp"
#include "trs/libs/transaction.hpp"
#include "trs/libs/function_manager.hpp"
#include "trs/libs/inspector.hpp"
#include "trs/libs/database.hpp"
#include "trs/helpers/verification.hpp"
#include "trs/helpers/execute.hpp"

namespace trs
{
  class server : public httplib::Server
  {
    public:
      server() = default;
      ~server() = default;

    public:
      void init(std::string service_name)
      {
        set_read_timeout(5, 0);
        set_write_timeout(5, 0);
        set_idle_interval(0, 1'000'000);
        set_payload_max_length(1024 * 1024 * 512); // 512MB
        set_CORS();
        set_post();
        m_service_name = service_name;
      }

      void set_log_connection(std::string connection) { database::instance().set_log_connection(connection); }
      void set_app_connection(std::string connection) { database::instance().set_app_connection(connection); }
      
      void log_to_console(bool v) { m_log_to_console = v; }

      template<typename Function>
      void registration(std::string func_name, Function insp, Function func) 
      { 
        m_inspector.registration(func_name, insp);
        m_function_manager.registration(func_name, func); 
      }

      void run()
      {
        try 
        {
          hack::log(" ")("server listen:", var::HOST, var::PORT);
          listen(var::HOST, var::PORT);
        }
        catch(std::exception& e)
        {
          hack::error(" ")("error server run:", e.what());
        }
        catch(...)
        {
          hack::error()("SUPPER ERROR!!! GOOD LUCK MY FRIEND!!! :)");
        }
      }

    private:
      std::string m_service_name;
      bool m_log_to_console { true };
      function_manager<transaction> m_function_manager;
      inspector<transaction> m_inspector;

    private:
      void set_CORS()
      {
        Options(R"(\*)", [](const auto& req, auto& res) { res.set_header("Allow", "POST, HEAD, OPTIONS"); });
        Options(var::API_URL, [](const auto& req, auto& res)
        {
          res.set_header("Access-Control-Allow-Origin", "*");
          res.set_header("Allow", "POST, HEAD, OPTIONS");
          res.set_header("Access-Control-Allow-Headers", "X-Requested-With, Content-Type, Accept, Origin, Authorization, TRS-server-function, TRS-server-token");
          res.set_header("Access-Control-Allow-Methods", "OPTIONS, HEAD, POST");
        });
      }
      
      void set_post()
      {
        Post(var::API_URL, [&](const httplib::Request& req, httplib::Response& res) {
          res.set_header("Access-Control-Allow-Origin", "*");
          res.set_header("Access-Control-Allow-Headers", "*");
          res.set_header("Access-Control-Allow-Methods", "POST");
          res.set_header("Access-Control-Allow-Credentials", "false");

          transaction tr;

          try
          {
            tr.set_data(req);
            verification(m_inspector, m_function_manager, tr);
            execute(m_function_manager, tr);
          }
          catch(hack::exception& ex)
          {
            ex.transaction(tr);
            ex.service(m_service_name);
            if (m_log_to_console)
              ex.log();
            ex.commit<database>();
          }

          res.set_content(nlohmann::to_string(tr.m_data.m_result), var::HEADER_FLAG_JSON); 
        });
      }
  };

  class client
  {
    public:
      client() = delete;
      client(std::string url) : m_cli { url } {}
      client(client&& c) : m_cli{ std::move(c.m_cli) }, m_token { std::move(c.m_token) } {}
      ~client() = default;

    public:
      void set_token(std::string t) { m_token = t; }
      auto post(std::string func_name, JSON&& data)
      {
        httplib::Headers headers = {
          { "TRS-server-token", m_token },
          { "TRS-server-function", func_name },
          { "Content-Type", "application/json" },
          { "Cookie", "unione_lang=ru" }
        };

        return m_cli.Post("/", headers, data.dump(),"application/json" );
      }

      auto post(std::string func_name, JSON& data)
      {
        httplib::Headers headers = {
          { "TRS-server-token", m_token },
          { "TRS-server-function", func_name },
          { "Content-Type", "application/json" },
          { "Cookie", "unione_lang=ru" }
        };

        return m_cli.Post("/", headers, data.dump(),"application/json" );
      }

    private:
      httplib::Client m_cli;
      std::string m_token;
  };
}


